import globals
import graphics

from OpenGL.GL import *

import time
import math

class Animation:
    '''
    Class for handling animations. This will load the texture and when rendered only render
    the correct sub-image . Animation parameters are loaded from a data file. The datafile
    specifies the texture image to use

    '''

    def __init__(self, filename, looping=False):
        f = open(filename, 'r')
        data = f.readlines()

        self.name = data[1].strip()
        self.imageFilename = data[3].strip()
        self.framerate = int(data[5].strip())
        self.tileWidth = int(data[7].strip())
        self.tileHeight = int(data[9].strip())
        self.frameWidth = int(data[11].strip())
        self.frameHeight = int(data[13].strip())

        self.frames = self.tileWidth * self.tileHeight

        self.texture = graphics.Texture2D(self.imageFilename)

        #Check that the tiling parameters match the image pixel sizes
        if self.tileWidth*self.frameWidth != self.texture.getPixelWidth() or \
            self.tileHeight*self.frameHeight != self.texture.getPixelHeight():

            print "Error loading animation %s. The data does not match the supplied image." % self.imageFilename
            print "  Subframe %d x %d and Tiling %d x %d ==> %d x %d. Image is %d x %d" % (
            self.frameWidth, self.frameHeight, self.tileWidth, self.tileHeight,
            self.frameWidth*self.tileWidth, self.frameHeight*self.tileHeight,
            self.texture.getPixelWidth(), self.texture.getPixelHeight())

            #exit()


        self.looping = looping

        self.playing = looping
        self.startTime = time.clock()

    def play(self):
        self.playing = True
        self.startTime = time.clock()

    def stop(self):
        self.playing = False

    def getFrameNumber(self):
        elapsed = time.clock()-self.startTime
        frameNum = elapsed * self.framerate
        return math.floor(frameNum+1) #Ceiling the value to get a whole frame number


    def draw(self, size=1, x=0, y=0):

        if not self.playing: return

        #Determine which frame we should be rendering using the 
        frameNumber = self.getFrameNumber()

        #Check that we aren't passed the end of the animation
        if frameNumber >= self.frames:
            if not self.looping:
                self.stop()
                return
            else:
                self.play()
                frameNumber = self.getFrameNumber()

        #Determine the texture coordinates for the current frame
        #The percentage of the texture accupied by a single animation frame
        xfrac = 1.0/self.tileWidth 
        yfrax = 1.0/self.tileHeight
        #The texture coordinates needed can be easily determined using the x/yfracs
        tl = frameNumber * xfrac 
        tr = (frameNumber+1) * xfrac
        tt = 1.0 #TODO This doesn't support properly tiled textures yet - only tiled in a single row
        tb = 0.0

        width=size  #TODO keep these in the proper aspect ratio
        height=float(size) * self.frameHeight / self.frameWidth

        glEnable(GL_TEXTURE_2D)
        glBindTexture(GL_TEXTURE_2D, self.texture.textureID)
        glColor3f(1, 1, 1)

        glPushMatrix()
        glTranslatef(x-(width/2), y-(height/2), 0)


        glBegin(GL_QUADS)

        glTexCoord2f(tl, tt)
        glVertex3f(0.0, 0.0, 0.0)

        glTexCoord2f(tr, tt)
        glVertex3f(width, 0.0, 0.0)

        glTexCoord2f(tr, tb)
        glVertex3f(width, height, 0.0)
        
        glTexCoord2f(tl, tb)
        glVertex3f(0.0, height, 0.0)
        glEnd()

        glDisable(GL_TEXTURE_2D)

        glPopMatrix()


    def drawDebug(self, size, x, y):

        width=size  
        height=size

        glPushMatrix()
        glTranslatef(x-(width/2), y-(height/2), 0)

        glBegin(GL_LINE_LOOP)

        glVertex3f(0.0, 0.0, 0.0)
        glVertex3f(width, 0.0, 0.0)
        glVertex3f(width, height, 0.0)
        glVertex3f(0.0, height, 0.0)

        glEnd()

        glPopMatrix()









