'''
BoatViz User Interface for The Bandit Naval warfare RTS game
'''

import globals
import graphics
import animation
import influencemap
import population
import individual
import vector

#import time


import math

from OpenGL.GL import *

MAXTIME = 600 # in seconds
USEGRAPHICS = True

class Entity:

    counter = 0
    drawInfluenceMap = False

    def __init__(self, filename, teamID):

        '''
        Creates a boat entity, loading the data from the supplied filename
        '''
        f = open(filename, 'r')
        data = f.readlines()

        self.name = data[1].strip()
        self.imagefilename = data[3].strip() #need to strip the trailing /n off
        self.length = int(data[5].strip())
        self.width = int(data[7].strip())
        self.flagPathCount = 0

        self.turningCircle = float(data[23].strip())

        boattype = filename[:-4] #use the filename and strip the last 4 chars
        self.__class__.counter += 1
        self.uniquename = boattype + `self.__class__.counter`

        self.teamID = teamID

        #leader flag
        self.leader = False

        #Starting location
        self.x = 0
        self.y = 0
        self.orientation = 0

        #Parameters for rendering
        self.selected = False
        self.selectionRingRotation = 0

        #History of position used to draw movement trails
        self.positionHistory = None

        self.influenceMap = influencemap.InfluenceMap(globals.scenario.terrainWidth, \
            globals.scenario.terrainHeight, 40, 39)

        self.alive = True

        self.trace = False


    def initGraphics(self, filename, teamID):

        '''
        Creates a boat entity, loading the data from the supplied filename
        '''
        f = open(filename, 'r')
        data = f.readlines()

        #Read the gun locations out - they are listed as pairs of x/y coords
        gunloc = data[35].split()
        self.gunLocations = []
        for i in range(0, len(gunloc), 2):
            self.gunLocations.append((int(gunloc[i]), int(gunloc[i+1]), \
            animation.Animation("data/explosion_1.dat") ) )

        self.gunAnimationSize = int(data[37].strip())

        #Parameters for rendering
        self.selected = False
        self.selectionRingRotation = 0

        #Texture to draw
        self.texture = graphics.Texture2D(self.imagefilename)

        if self.trace:
            print "Loaded Boat: %s : %s. Boat dimensions: %d x %d" % (self.name, self.imagefilename, self.length, self.width) 

        self.explosion = animation.Animation("data/explosion_1_slow.dat")

        #History of position used to draw movement trails
        self.positionHistory = None

    def setDestination(self, x, y):
        waypoints = self.influenceMap.findPath(self.x, self.y, x, y)

        globals.gameEngine.setDestination(self.uniquename, self.x, self.y, 0)
        for i in waypoints:
            globals.gameEngine.addDestination(self.uniquename, i[0], i[1], 0)

    def setWaypoint(self, x, y):
        globals.gameEngine.setDestination(self.uniquename, self.x, self.y, 0)
        globals.gameEngine.addDestination(self.uniquename, x, y, 0)        
            

    def addDestination(self, x, y):
        globals.gameEngine.addDestination(self.uniquename, x, y, 0)
        


    def draw(self):

        if not self.alive:
            return

        self.x, self.y = globals.gameEngine.getPosition(self.uniquename)
        self.orientation = globals.gameEngine.getOrientation(self.uniquename)

        glPushMatrix()


        glTranslatef(self.x, self.y, 0)
        glRotatef(-self.orientation, 0, 0, 1)


        glTranslatef(-self.width/2, -self.length/2, 0)
        
        self.texture.draw(self.width, self.length)

        glPopMatrix()

        explosionSize = min(self.width, self.length) * 10 
        self.explosion.draw(explosionSize, self.x, self.y)

        #Draw the explosions for the guns firing. The positions are relative to the center of the
        #boat (unrotated) so need the opengl to help place them appropriately
        glPushMatrix()
        glTranslatef(self.x, self.y, 0)
        glRotatef(-self.orientation, 0, 0, 1)
        for i in self.gunLocations:
            gx, gy, ganim = i
            ganim.draw(self.gunAnimationSize, gx, gy)
        glPopMatrix()

        #Draw (if approriate) lines describing the target, and/or destination of a boat
        dst = globals.gameEngine.getDestinations(self.uniquename)
        if dst:
            lastPoint = (self.x, self.y)
            for i in dst:
                graphics.drawLine(lastPoint[0], lastPoint[1], i[0], i[1], EntityManager.teamColors[self.teamID])
                graphics.drawCircle(i[0], i[1], radius=10, color=EntityManager.teamColors[self.teamID])
                lastPoint = i

        '''
        dst = globals.gameEngine.getTargetPosition(self.uniquename)
        if dst:
            graphics.drawLine(self.x, self.y, dst[0], dst[1], EntityManager.teamColors[self.teamID])
        '''

        #Draw a circle around the entity 10% bigger than its length
        radius = max(self.width/2, self.length/2) * 1.1
        lineWidth = 1
        if self.selected:
            lineWidth = 5

        #Rotate the selection ring if the entity is currently selected. Becuase it looks pretty 
        #if its spinning. Ring is sput at 90degrees per second
        if self.selected: self.selectionRingRotation += 90*globals.dt

        if self.leader:
            graphics.drawSelectionRing(self.x, self.y, radius, EntityManager.teamColors[3], self.selected, self.selectionRingRotation)
        else:    
            graphics.drawSelectionRing(self.x, self.y, radius, EntityManager.teamColors[self.teamID], self.selected, self.selectionRingRotation)

        #Draw the health bar below the boat
        health = globals.gameEngine.getHealth(self.uniquename)

        radius *= 1.4  #Radius needs to be increased to account for the selection ring
        graphics.drawRectangle(self.x-50, self.y-radius-20, 200, 15, \
            EntityManager.teamColors[self.teamID], filled=False)

        graphics.drawRectangle(self.x-50, self.y-radius-20, 200*health, 15, \
            EntityManager.teamColors[self.teamID], filled=True)

        #Draw the information about the influence map
        if self.selected and Entity.drawInfluenceMap:
            self.influenceMap.drawDebug()

    def drawDebug(self):

        #Initialise the history if this is the first time using it
        if not self.positionHistory:
            self.positionHistory = []
            self.positionHistory.append((self.x, self.y))

        self.positionHistory.append((self.x, self.y))

        #Maintain a rolling history of only 1000 points
        if len(self.positionHistory) > 1000: 
            del self.positionHistory[0]

        glColor3f(1, 0, 0)
        glBegin(GL_LINE_STRIP)
        for i in self.positionHistory:
            glVertex(i[0], i[1], 0)    
        glEnd()


        #Draw turning circles
        glPushMatrix()
        glTranslatef(self.x, self.y, 0)
        glRotatef(-self.orientation, 0, 0, 1)
        glTranslatef(self.turningCircle, 0, 0)
        graphics.drawCircle(0, 0, self.turningCircle, color=(0, 0, 1))
        glTranslatef(-2.0*self.turningCircle, 0, 0)
        graphics.drawCircle(0, 0, self.turningCircle, color=(0, 0, 1))
        glPopMatrix()



    def hit(self):
        self.explosion.play()
        globals.gameEngine.hit(self.uniquename)
        b = globals.gameEngine.findBoat(self.uniquename)

    def fireGuns(self):
        for i in self.gunLocations:
            i[2].play()
    
    def explode(self):
        self.explosion.play()

    def destroy(self):
        if USEGRAPHICS:
            self.explosion.play()
        self.alive = False


class Projectile:
    
    def __init__(self, name, width, length):
        self.uniquename = name
        self.width = width
        self.length = length

        self.x = 0
        self.y = 0
        self.orientation = 0

    def draw(self):
        self.x, self.y = globals.gameEngine.getPosition(self.uniquename)
        self.orientation = globals.gameEngine.getOrientation(self.uniquename)

        glPushMatrix()
        glColor3f(0, 0, 1)
        glTranslatef(self.x, self.y, 0)
        glRotatef(-self.orientation, 0, 0, 1)
        glTranslatef(-self.width/2, -self.length/2, 0)
        glBegin(GL_TRIANGLES)
        glVertex2f(0, 0)
        glVertex2f(self.width, 0)
        glVertex2f(self.width/2, self.length)
        glEnd()
        glPopMatrix()



class EntityManager:
    
    teamColors = [(0, 0, 0), (0, 1, 0), (1, 0, 0), (0, 0, 1)]

    def __init__(self, indiv):
        self.win = None
        self.entities = []
        self.projectiles = []
        self.ga = indiv 
        self.currenttime= 0

        self.attacker = indiv
        self.defender = None

    def initCoevolve(self, defender):
        self.defender= defender

    def gameOver(self):
        if self.win != None:
            return True
        return False

    def timeLeft(self):
        global MAXTIME
        return MAXTIME - self.currenttime

    def update(self):

        if self.win != None:
            print "Game End"
            return False;
        else: 
            self.currenttime += globals.dt
            e = globals.gameEngine.getEvent()
            while e != None:
                if e[0] == "WEAPON_ANIMATION":
                    if not self.findBoatByName(e[1]) == None:
                        self.findBoatByName(e[1]).fireGuns()
                if e[0] == "ADD_PROJECTILE":
                    self.projectiles.append(Projectile(e[1], e[2], e[3]))
                if e[0] == "REMOVE_PROJECTILE":
                    for i in self.projectiles:
                        if i.uniquename == e[1]:
                            self.projectiles.remove(i)
                            break;
                if e[0] == "EXPLODE_BOAT":
                    if not self.findBoatByName(e[1]) == None:
                        self.findBoatByName(e[1]).explode()
                if e[0] == "DESTROY_BOAT":
                    if not self.findBoatByName(e[1]) == None and self.findBoatByName(e[1]).alive:
                        self.findBoatByName(e[1]).destroy()

                e = globals.gameEngine.getEvent()

        self.generateMap()
        return True

    def generateMap(self):
        #Generate the maps to run the GA with
        #TODO - this is very hacky and makes lots of assumptions about how many boats there are
        boatmap = []
        for i in range(self.entities[0].influenceMap.gridWidth):
            boatmap.append([0]*self.entities[0].influenceMap.gridHeight)

        #find all the attacker boats
        attackerboats = []
        defenderboats= []
        for i in self.entities:
            if i.teamID == 1:
                attackerboats.append(i)
            if i.teamID == 2:
                defenderboats.append(i)
        
        #assign the leader
		'''
        count = 0
        for i in attackerboats:
            if count == 0:
                i.leader = True
            else:
                i.leader = False    
            count +=1    
		'''

        #Make sure we have as many boats as the GA expects
        if not len(attackerboats) <= individual.FRIENDLY_NUM or not len(defenderboats) <= individual.ENEMY_NUM:
            print "There are an incorrect number of attacker or enemy boats. attacker %d != %d. Enemy \
            %d != %d" % (len(attackerboats), individual.FRIENDLY_NUM, len(defenderboats), individual.ENEMY_NUM)
            exit(0)

        #Add the index of each boat into the map 1-5 for friendly, 6-10 for defender
        for i in attackerboats:
            x, y = i.influenceMap.worldToGrid(i.x, i.y)
            if i.alive:
                boatmap[x][y] = attackerboats.index(i)+1

        for i in defenderboats:
            x, y = i.influenceMap.worldToGrid(i.x, i.y)
            if i.alive:
                boatmap[x][y] = defenderboats.index(i)+6


        flagx, flagy = attackerboats[0].influenceMap.worldToGrid(globals.scenario.flagx, globals.scenario.flagy)
        boatmap[flagx][flagy] = -1

        # if coevolving, set the IMs for the defenderboats
        if globals.COEVOLVE:
            influenceMaps = self.attacker.generateMaps(boatmap)
            defenderInfluenceMaps= self.defender.generateMaps(boatmap)
            
            for i in defenderboats:
                i.influenceMap.setWeights(defenderInfluenceMaps[defenderboats.index(i)])
        else:
            influenceMaps = self.ga.generateMaps(boatmap)

        for i in attackerboats:
            i.influenceMap.setWeights(influenceMaps[attackerboats.index(i)])

        # check for end game state
        deadcount = 0
        for i in attackerboats:
            tempVector = vector.Vector(globals.scenario.flagx - i.x, globals.scenario.flagy - i.y)
            if tempVector.length() < 200:
                #flag reached
                self.win = True
                break
            if not i.alive:
                deadcount += 1
        # all entities dead
        if deadcount == individual.FRIENDLY_NUM:
            self.win = False
        # time ended
        global MAXTIME
        #if (time.clock() - self.currenttime) > MAXTIME:
        if self.currenttime > MAXTIME:
            self.win = False

        #Auto target finding
        if globals.scenario.autoAttack:
            for i in self.entities:
                if not i.alive:
                    continue
                closest = None
                dist = -1
                for j in self.entities:
                    if not j.alive or j.teamID == i.teamID:
                        continue
                    d = vector.Vector(i.x-j.x, i.y-j.y).length()
                    if closest == None or d<dist:
                        closest = j
                        dist = d

                if closest:
                    globals.gameEngine.attack(i.uniquename, closest.uniquename)
       

        #For each attacker boat find the lowest weight on the GA map and route there
        #only do this if the leader is still alive
        leaderalive = True
        for i in attackerboats:
            if i.leader:
                if not i.alive:
                    leaderalive = False
                break

        if globals.scenario.autoMove and leaderalive:
            for i in attackerboats:
                if i.alive:
                    x, y = i.influenceMap.findLowestWeight(globals.scenario)
                    followingFriend = False
                    
                    # check if the boat is following a friend
                    for j in attackerboats:
                        fx, fy = j.influenceMap.worldToGrid(j.x, j.y)
                        ix, iy = i.influenceMap.worldToGrid(x, y)
                        if j.alive == True and fx == ix and fy == iy:
                            followingFriend = True
                    #Make sure we aren't already heading to the same point to prevent 
                    #unnessesary A* runs which significanly slows the game down

					
                    destinations = globals.gameEngine.getDestinations(i.uniquename)

                    # if the waypoints are empty	
                    if not destinations or destinations == None:
                        tempvec = vector.Vector(i.x - x, i.y-y)
                        if tempvec.length() < 100:
                            i.setWaypoint(x, y)
                        else: 
                            i.setDestination(x, y)
                    else:

                        # if the destination is the flag, recalculate path every so often
                        oldDestWorldx, oldDestWorldy = destinations[-1]
                        flagx, flagy = i.influenceMap.worldToGrid(globals.scenario.flagx, globals.scenario.flagy)
                        destx, desty = i.influenceMap.worldToGrid(oldDestWorldx, oldDestWorldy)
                        if flagx == destx and flagy == desty:
                            if i.flagPathCount < 50:
                                i.flagPathCount += 1
                            else:
                                i.setDestination(x, y)
                                i.flagPathCount = 0
                        else:    

                            # if the destination has moved more than 200 metres, recalculate
                            distvec = vector.Vector(oldDestWorldx - x, oldDestWorldy - y)
                            if distvec.length() > 200:
                                i.setDestination(x, y)

                    '''
                    elif followingFriend:
                        dx, dy = destinations[-1]
                        distvec = vector.Vector(x-dx, y-dy)
                        if distvec.length() > 200:
                            i.setDestination(x, y)
                    else:
                        # check if heading to defender, and if defender still exists
                        dx, dy = destinations[-1]
                        for j in defenderboats:
                            jx, jy = i.influenceMap.worldToGrid(dx, dy)
                            if dx == jx and dy == jy and not i.alive:
                                i.setDestination(x, y)
                    '''

            # if coevolving, move defender boats too
            if globals.COEVOLVE:
                for i in defenderboats:
                    if i.alive:
                        x, y = i.influenceMap.findLowestWeight(globals.scenario)
                        followingFriend = False
                        
                        # check if the boat is following a friend
                        for j in defenderboats:
                            fx, fy = j.influenceMap.worldToGrid(j.x, j.y)
                            ix, iy = i.influenceMap.worldToGrid(x, y)
                            if j.alive == True and fx == ix and fy == iy:
                                followingFriend = True
                        #Make sure we aren't already heading to the same point to prevent 
                        #unnessesary A* runs which significanly slows the game down
                        destinations = globals.gameEngine.getDestinations(i.uniquename)
                        if not destinations or destinations == None:
                            tempvec = vector.Vector(i.x - x, i.y-y)
                            if tempvec.length() < 100:
                                i.setWaypoint(x, y)
                            else: 
                                i.setDestination(x, y)
                        else:
                            oldDestWorldx, oldDestWorldy = destinations[-1]

                            # if the destination has moved more than 200 metres, recalculate
                            distvec = vector.Vector(oldDestWorldx - x, oldDestWorldy - y)
                            if distvec.length() > 200:
                                i.setDestination(x, y)
                            


    def debug(self):
        print "Printing entitymanager debug - num enditites %d" % len(self.entities)

    def moveEntities(self):
        for i in self.entities:
            if not i.alive:
                continue
            i.x, i.y = globals.gameEngine.getPosition(i.uniquename)
            i.orientation = globals.gameEngine.getOrientation(i.uniquename)
        
        for i in self.projectiles:
            i.x, i.y = globals.gameEngine.getPosition(i.uniquename)
            i.orientation = globals.gameEngine.getOrientation(i.uniquename)

    def renderEntities(self):
        for i in self.entities:
            i.draw()
        for i in self.projectiles:
            i.draw()

    
    def addAircraftCarrier(self, teamID, x=0, y=0):
        newent = Entity("data/aircraftcarrier.dat", teamID)
        self.entities.append(newent)
        newent.x = x
        newent.y = y
        globals.gameEngine.makeAircraftCarrier(newent.uniquename, x, y, 0)
        return newent

    def addDestroyer(self,teamID, x=0, y=0):
        newent = Entity("data/destroyer.dat", teamID)
        self.entities.append(newent) 
        newent.x = x
        newent.y = y
        globals.gameEngine.makeDestroyer(newent.uniquename, x, y, 0)
        return newent

    def addCargoship(self, teamID, x=0, y=0):
        newent = Entity("data/cargoship.dat", teamID)
        self.entities.append(newent)
        newent.x = x
        newent.y = y
        globals.gameEngine.makeCargoship(newent.uniquename, x, y, 0)
        return newent


    def findBoatByName(self, name):
        for i in self.entities:
            if i.uniquename == name:
                return i
        return None


    def findBoatAt(self, x, y):
        '''
        Determine which object is positioned at the given coordinates
        '''

        found = None

        #For each boat we compare the click against the center point, and maintain which entity
        #was closest to the mouseclick
        for i in self.entities:
            if not i.alive:
                continue
            v = ((i.x-x) * (i.x-x)) + ((i.y-y) * (i.y-y))
            dst = math.sqrt(v)
            if found == None:
                found = (i, dst)
            elif dst < found[1]:
                found = (i, dst)

        #Check to make sure the click was close enough to the boat to count as a selection
        #Here we just use the circle that fits around the longest dimension
        if found[1] < max(found[0].length/2, found[0].width/2) * 1.4:
            return found[0]
        else:
            return None


    def pointSelect(self, x, y):
        '''
        This is called for a left-click action - we select a boat if one was under the cursor, or 
        deselect all if they clicked nowhere
        '''
        #Deselect all boats
        for i in self.entities:
            i.selected = False

        clicked = self.findBoatAt(x, y)
        if clicked != None:
            clicked.selected = True
        
    
    def pointAltSelect(self, x, y):
        '''
        Clicked with shift held down, for adding units to the select group
        '''
        clicked = self.findBoatAt(x, y)
        #if clicked != None and clicked.teamID == 1:
        if clicked != None:
            clicked.selected = True

    def areaSelect(self, region):
        '''Select everything in a region'''     

        #TODO: Work out how region (passed from ocean.py) can ever be null, it shouldn't be possible - but is!
        if not region: return

        #Deselect all entities first
        for i in self.entities:
            i.selected = False

        self.areaAltSelect(region)

    def areaAltSelect(self, region):
        '''Find all boats that lie within the region specified [(x1, y1), (x2, y2)]'''
        #TODO: Work out how region (passed from ocean.py) can ever be null, it shouldn't be possible - but is!
        if not region: return

        left = min(region[0][0], region[1][0])
        right = max(region[0][0], region[1][0])
        top = max(region[0][1], region[1][1])
        bottom = min(region[0][1], region[1][1])
        for i in self.entities:
            #if left <= i.x <= right and bottom <= i.y <= top and i.teamID == 1 and i.alive:
            if left <= i.x <= right and bottom <= i.y <= top and i.alive:
                i.selected = True

        #Becuase this gets called for any single clicks, we need to make sure we are including any
        #entities that were 'near' the start and final mouse clicks otherwise things don't work
        #properly
        self.pointAltSelect(region[0][0], region[0][1])
        self.pointAltSelect(region[1][0], region[1][1])

    def pointAction(self, x, y):
        '''
        Action command, this specifies that the selected objects should move, or attack (depeding
        on what they clicked on
        '''
        actioned = self.findBoatAt(x, y)
        if actioned == None:
            #Probably a move command
            for i in self.entities:
                if i.selected:
                    i.setDestination(x, y)
        else:
            #Might be an attack command
            #First work out if we have our own units selected
            for i in self.entities:
                #if i.selected and i.teamID == 1:
                if i.selected:
                    globals.gameEngine.attack(i.uniquename, actioned.uniquename)

    def pointAltAction(self, x, y):
        '''
        Alt action command (shift held down) - We add a waypoint rather than just setting the
        destination
        '''
        actioned = self.findBoatAt(x, y)
        if not actioned:
            for i in self.entities:
                if i.selected:
                    i.addDestination(x, y)
                




    def pointDebug(self, x, y):
        newent = self.addDestroyer(1)
        newent.x = x
        newent.y = y

